@@ -58,10 +58,10 @@
  * The format record for the GIF file format:
  */
 
-static int      ChanMatchGIF _ANSI_ARGS_((Tcl_Interp *interp, Tcl_Channel chan, struct Tcl_Obj *fileName,
-		    struct Tcl_Obj *format, int *widthPtr, int *heightPtr));
-static int	ObjMatchGIF _ANSI_ARGS_((Tcl_Interp *interp, struct Tcl_Obj *data,
-		    struct Tcl_Obj *format, int *widthPtr, int *heightPtr));
+static int      ChanMatchGIF _ANSI_ARGS_((Tcl_Channel chan, struct Tcl_Obj *fileName,
+		    struct Tcl_Obj *format, int *widthPtr, int *heightPtr, Tcl_Interp *interp));
+static int	ObjMatchGIF _ANSI_ARGS_((struct Tcl_Obj *data,
+		    struct Tcl_Obj *format, int *widthPtr, int *heightPtr, Tcl_Interp *interp));
 static int      ChanReadGIF  _ANSI_ARGS_((Tcl_Interp *interp,
 		    Tcl_Channel chan, struct Tcl_Obj *fileName, struct Tcl_Obj *format,
 		    Tk_PhotoHandle imageHandle, int destX, int destY,
@@ -74,8 +74,7 @@ static int 	ChanWriteGIF _ANSI_ARGS_(( Tcl_Interp *interp,
 		    char *filename, Tcl_Obj *format,
 		    Tk_PhotoImageBlock *blockPtr));
 static int	StringWriteGIF _ANSI_ARGS_((Tcl_Interp *interp,
-		    Tcl_DString *dataPtr, Tcl_Obj *format,
-		    Tk_PhotoImageBlock *blockPtr));
+		    Tcl_Obj *format,  Tk_PhotoImageBlock *blockPtr));
 static int      CommonReadGIF  _ANSI_ARGS_((Tcl_Interp *interp,
 		    MFile *handle, CONST char *fileName, Tcl_Obj *format,
 		    Tk_PhotoHandle imageHandle, int destX, int destY,
@@ -126,7 +125,7 @@ static int		ReadImage _ANSI_ARGS_((Tcl_Interp *interp,
 			    int width, int height, int srcX, int srcY,
 			    int interlace, int transparent));
 
-
+
 /*
  *----------------------------------------------------------------------
  *
@@ -146,14 +145,14 @@ static int		ReadImage _ANSI_ARGS_((Tcl_Interp *interp,
  */
 
 static int
-ChanMatchGIF(interp, chan, fileName, format, widthPtr, heightPtr)
-    Tcl_Interp *interp;
+ChanMatchGIF(chan, fileName, format, widthPtr, heightPtr, interp)
     Tcl_Channel chan;		/* The image channel, open for reading. */
     struct Tcl_Obj *fileName;	/* The name of the image file. */
     struct Tcl_Obj *format;	/* User-specified format object, or NULL. */
     int *widthPtr, *heightPtr;	/* The dimensions of the image are
 				 * returned here if the file is a valid
 				 * raw GIF file. */
+    Tcl_Interp *interp;
 {
     MFile handle;
 
@@ -164,7 +163,7 @@ ChanMatchGIF(interp, chan, fileName, format, widthPtr, heightPtr)
 
     return ReadGIFHeader(&handle, widthPtr, heightPtr);
 }
-
+
 /*
  *----------------------------------------------------------------------
  *
@@ -484,7 +483,7 @@ CommonReadGIF(interp, handle, fileName, format, imageHandle, destX, destY,
     }
 
     if (transparent == -1) {
-	Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, height);
+	Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, height, TK_PHOTO_COMPOSITE_SET);
     } else {
 	ImgPhotoPutBlock(imageHandle, &block, destX, destY, width, height);
     }
@@ -502,7 +501,7 @@ CommonReadGIF(interp, handle, fileName, format, imageHandle, destX, destY,
     return TCL_ERROR;
 
 }
-
+
 /*
  *----------------------------------------------------------------------
  *
@@ -522,12 +521,12 @@ CommonReadGIF(interp, handle, fileName, format, imageHandle, destX, destY,
  */
 
 static int
-ObjMatchGIF(interp, data, format, widthPtr, heightPtr)
-    Tcl_Interp *interp;		/* interpreter */
+ObjMatchGIF(data, format, widthPtr, heightPtr, interp)
     Tcl_Obj *data;		/* the object containing the image data */
     Tcl_Obj *format;		/* the image format object */
     int *widthPtr;		/* where to put the image width */
     int *heightPtr;		/* where to put the image height */
+    Tcl_Interp *interp;		/* interpreter */
 {
     MFile handle;
 
@@ -538,7 +537,7 @@ ObjMatchGIF(interp, data, format, widthPtr, heightPtr)
     }
     return ReadGIFHeader(&handle, widthPtr, heightPtr);
 }
-
+
 /*
  *----------------------------------------------------------------------
  *
@@ -641,7 +640,7 @@ ReadColorMap(handle, number, buffer)
 	    if (! ReadOK(handle, rgb, sizeof(rgb))) {
 		return 0;
 	    }
-	
+
 	    if (buffer) {
 		buffer[i][CM_RED] = rgb[0] ;
 		buffer[i][CM_GREEN] = rgb[1] ;
@@ -666,7 +665,7 @@ DoExtension(handle, label, transparent)
     switch (label) {
 	case 0x01:      /* Plain Text Extension */
 	    break;
-	
+
 	case 0xff:      /* Application Extension */
 	    break;
 
@@ -720,7 +719,7 @@ GetDataBlock(handle, buf)
 }
 
 
-
+
 /*
  *----------------------------------------------------------------------
  *
@@ -777,7 +776,7 @@ ReadImage(interp, imagePtr, handle, len, rows, cmap,
     /*
      *  Initialize the decoder
      */
-    if (! ReadOK(handle,&initialCodeSize,1))  {
+    if (! ReadOK(handle,&initialCodeSize,1) || inititalCodeSize > MAX_LWZ_BITS)  {
 	Tcl_AppendResult(interp, "error reading GIF image: ",
 		Tcl_PosixError(interp), (char *) NULL);
 	return TCL_ERROR;
@@ -844,7 +843,7 @@ ReadImage(interp, imagePtr, handle, len, rows, cmap,
 		    oldCode     = -1;
 		    continue;
 		}
-		
+
 		if (oldCode == -1) {
 		    /*
 		     * Last pass reset the decoder, so the first code we
@@ -859,7 +858,7 @@ ReadImage(interp, imagePtr, handle, len, rows, cmap,
 		    firstCode = code;
 		    continue;
 		}
-		
+
 		inCode = code;
 
 		if (code == maxCode) {
@@ -1055,7 +1054,7 @@ GetCode(handle, code_size, flag)
 
 
 
-
+
 /*
  * ChanWriteGIF - writes a image in GIF format.
  *-------------------------------------------------------------------------
@@ -1087,7 +1086,7 @@ GetCode(handle, code_size, flag)
   *  Types, defines and variables needed to write and compress a GIF.
   */
 
-typedef int (* ifunptr) _ANSI_ARGS_((void));	
+typedef int (* ifunptr) _ANSI_ARGS_((void));
 
 #define LSB(a)                  ((unsigned char) (((short)(a)) & 0x00FF))
 #define MSB(a)                  ((unsigned char) (((short)(a)) >> 8))
@@ -1148,15 +1147,15 @@ ChanWriteGIF (interp, filename, format, blockPtr)
 }
 
 static int
-StringWriteGIF(interp, dataPtr, format, blockPtr)
+StringWriteGIF(interp, format, blockPtr)
     Tcl_Interp *interp;
-    Tcl_DString *dataPtr;
     Tcl_Obj *format;
     Tk_PhotoImageBlock *blockPtr;
 {
     int result;
     MFile handle;
     Tcl_DString data;
+    Tcl_DString *dataPtr;
 
     ImgFixStringWriteProc(&data, &interp, &dataPtr, &format, &blockPtr);
 
@@ -1292,7 +1291,7 @@ CommonWriteGIF(interp, handle, format, blockPtr)
     c = GIF_TERMINATOR;
     ImgPutc(c,handle);
 
-    return TCL_OK;	
+    return TCL_OK;
 }
 
 static int
@@ -1311,7 +1310,7 @@ color(red, green, blue)
     return -1;
 }
 
-
+
 static int
 nuevo(red, green, blue, mapa)
     int red,green,blue;
@@ -1410,7 +1409,7 @@ int colors;
     return (bits-1);
 }
 
-
+
 
 /*-----------------------------------------------------------------------
  *
@@ -1728,7 +1727,7 @@ static void rl_flush_withtable(count)
     repleft = 1 + compute_triangle_count(leftover,max_ocodes);
   }
  if (VERBOSE) printf("rl_flush_withtable repmax=%d leftover=%d repleft=%d\n",repmax,leftover,repleft);
- if (1+compute_triangle_count(count,max_ocodes) < repmax+repleft)
+ if (1+(int) compute_triangle_count(count,max_ocodes) < repmax+repleft)
   { output(code_clear);
     did_clear();
     rl_flush_fromclear(count);
@@ -1826,3 +1825,4 @@ static void compress( init_bits, handle, readValue )
  * End of miGIF section  - See copyright notice at start of section.
  *
  *-----------------------------------------------------------------------*/
+
