@@ -4,12 +4,12 @@
  *	This file implements text items for canvas widgets.
  *
  * Copyright (c) 1991-1994 The Regents of the University of California.
- * Copyright (c) 1994-1995 Sun Microsystems, Inc.
+ * Copyright (c) 1994-1997 Sun Microsystems, Inc.
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tkCanvText.c,v 1.3 1998/10/16 00:46:19 rjohnson Exp $
+ * RCS: @(#) $Id: tkCanvText.c,v 1.15 2003/02/09 07:48:22 hobbs Exp $
  */
 
 #include "tkPort.h"
@@ -35,17 +35,14 @@ typedef struct TextItem  {
      */
 
     double x, y;		/* Positioning point for text. */
-    int insertPos;		/* Insertion cursor is displayed just to left
-				 * of character with this index. */
+    int insertPos;		/* Character index of character just before
+				 * which the insertion cursor is displayed. */
 
     /*
      * Configuration settings that are updated by Tk_ConfigureWidget.
      */
 
     Tk_Anchor anchor;		/* Where to anchor text relative to (x,y). */
-    Tk_Tile tile;		/* Tile for text. */
-    Tk_Tile activeTile;		/* Tile for text if state is active. */
-    Tk_Tile disabledTile;	/* Tile for text if state is disabled. */
     Tk_TSOffset tsoffset;
     XColor *color;		/* Color for text. */
     XColor *activeColor;	/* Color for text. */
@@ -64,7 +61,8 @@ typedef struct TextItem  {
      * configuration settings above.
      */
 
-    int numChars;		/* Number of non-NULL characters in text. */
+    int numChars;		/* Length of text in characters. */
+    int numBytes;		/* Length of text in bytes. */
     Tk_TextLayout textLayout;	/* Cached text layout information. */
     int leftEdge;		/* Pixel location of the left edge of the
 				 * text item; where the left border of the
@@ -78,6 +76,7 @@ typedef struct TextItem  {
 				 * to use to draw the insertion cursor when
 				 * it's off.  Used if the selection and
 				 * insertion cursor colors are the same.  */
+    Tk_Tile tile;
 } TextItem;
 
 /*
@@ -85,8 +84,8 @@ typedef struct TextItem  {
  */
 
 static Tk_CustomOption stateOption = {
-    Tk_StateParseProc,
-    Tk_StatePrintProc, (ClientData) 2
+    TkStateParseProc,
+    TkStatePrintProc, (ClientData) 2
 };
 static Tk_CustomOption tagsOption = {
     Tk_CanvasTagsParseProc,
@@ -106,9 +105,6 @@ static Tk_ConfigSpec configSpecs[] = {
 	(char *) NULL, Tk_Offset(TextItem, activeColor), TK_CONFIG_NULL_OK},
     {TK_CONFIG_BITMAP, "-activestipple", (char *) NULL, (char *) NULL,
 	(char *) NULL, Tk_Offset(TextItem, activeStipple), TK_CONFIG_NULL_OK},
-    {TK_CONFIG_CUSTOM, "-activetile", (char *) NULL, (char *) NULL,
-	(char *) NULL, Tk_Offset(TextItem, activeTile),
-	TK_CONFIG_NULL_OK, &tileOption},
     {TK_CONFIG_ANCHOR, "-anchor", (char *) NULL, (char *) NULL,
 	"center", Tk_Offset(TextItem, anchor),
 	TK_CONFIG_DONT_SET_DEFAULT},
@@ -116,9 +112,6 @@ static Tk_ConfigSpec configSpecs[] = {
 	(char *) NULL, Tk_Offset(TextItem, disabledColor), TK_CONFIG_NULL_OK},
     {TK_CONFIG_BITMAP, "-disabledstipple", (char *) NULL, (char *) NULL,
 	(char *) NULL, Tk_Offset(TextItem, disabledStipple), TK_CONFIG_NULL_OK},
-    {TK_CONFIG_CUSTOM, "-disabledtile", (char *) NULL, (char *) NULL,
-	(char *) NULL, Tk_Offset(TextItem, disabledTile),
-	TK_CONFIG_NULL_OK, &tileOption},
     {TK_CONFIG_COLOR, "-fill", (char *) NULL, (char *) NULL,
 	"black", Tk_Offset(TextItem, color), TK_CONFIG_NULL_OK},
     {TK_CONFIG_FONT, "-font", (char *) NULL, (char *) NULL,
@@ -136,9 +129,6 @@ static Tk_ConfigSpec configSpecs[] = {
 	(char *) NULL, Tk_Offset(TextItem, stipple), TK_CONFIG_NULL_OK},
     {TK_CONFIG_CUSTOM, "-tags", (char *) NULL, (char *) NULL,
 	(char *) NULL, 0, TK_CONFIG_NULL_OK, &tagsOption},
-    {TK_CONFIG_CUSTOM, "-tile", (char *) NULL, (char *) NULL,
-	(char *) NULL, Tk_Offset(TextItem, tile),
-	TK_CONFIG_NULL_OK, &tileOption},
     {TK_CONFIG_STRING, "-text", (char *) NULL, (char *) NULL,
 	"", Tk_Offset(TextItem, text), 0},
     {TK_CONFIG_CALLBACK, "-updatecommand", (char *) NULL, (char *) NULL,
@@ -157,10 +147,10 @@ static void		ComputeTextBbox _ANSI_ARGS_((Tk_Canvas canvas,
 			    TextItem *textPtr));
 static int		ConfigureText _ANSI_ARGS_((Tcl_Interp *interp,
 			    Tk_Canvas canvas, Tk_Item *itemPtr, int argc,
-			    char **argv, int flags));
+			    Tcl_Obj *CONST objv[], int flags));
 static int		CreateText _ANSI_ARGS_((Tcl_Interp *interp,
 			    Tk_Canvas canvas, struct Tk_Item *itemPtr,
-			    int argc, char **argv));
+			    int argc, Tcl_Obj *CONST objv[]));
 static void		DeleteText _ANSI_ARGS_((Tk_Canvas canvas,
 			    Tk_Item *itemPtr, Display *display));
 static void		DisplayCanvText _ANSI_ARGS_((Tk_Canvas canvas,
@@ -179,7 +169,7 @@ static void		SetTextCursor _ANSI_ARGS_((Tk_Canvas canvas,
 			    Tk_Item *itemPtr, int index));
 static int		TextCoords _ANSI_ARGS_((Tcl_Interp *interp,
 			    Tk_Canvas canvas, Tk_Item *itemPtr,
-			    int argc, char **argv));
+			    int argc, Tcl_Obj *CONST objv[]));
 static void		TextDeleteChars _ANSI_ARGS_((Tk_Canvas canvas,
 			    Tk_Item *itemPtr, int first, int last));
 static void		TextInsert _ANSI_ARGS_((Tk_Canvas canvas,
@@ -221,10 +211,10 @@ Tk_ItemType tkTextType = {
     (Tk_ItemType *) NULL,		/* nextPtr */
     (Tk_ItemBboxProc *) ComputeTextBbox,/* bboxProc */
     Tk_Offset(Tk_VisitorType, visitText), /* acceptProc */
-    (Tk_ItemGetCoordProc *) NULL,	/* getCoordProc */
-    (Tk_ItemSetCoordProc *) NULL	/* setCoordProc */
+    NULL,	/* getCoordProc */
+    NULL	/* setCoordProc */
 };
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -236,7 +226,7 @@ Tk_ItemType tkTextType = {
  * Results:
  *	A standard Tcl return value.  If an error occurred in
  *	creating the item then an error message is left in
- *	interp->result;  in this case itemPtr is left uninitialized
+ *	the interp's result;  in this case itemPtr is left uninitialized
  *	so it can be safely freed by the caller.
  *
  * Side effects:
@@ -246,39 +236,24 @@ Tk_ItemType tkTextType = {
  */
 
 static int
-CreateText(interp, canvas, itemPtr, argc, argv)
-    Tcl_Interp *interp;			/* Interpreter for error reporting. */
-    Tk_Canvas canvas;			/* Canvas to hold new item. */
-    Tk_Item *itemPtr;			/* Record to hold new item;  header
-					 * has been initialized by caller. */
-    int argc;				/* Number of arguments in argv. */
-    char **argv;			/* Arguments describing rectangle. */
+CreateText(interp, canvas, itemPtr, objc, objv)
+    Tcl_Interp *interp;		/* Interpreter for error reporting. */
+    Tk_Canvas canvas;		/* Canvas to hold new item. */
+    Tk_Item *itemPtr;		/* Record to hold new item; header has been
+				 * initialized by caller. */
+    int objc;			/* Number of arguments in objv. */
+    Tcl_Obj *CONST objv[];	/* Arguments describing rectangle. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
     int i;
 
-    if (argc==1) {
-	i = 1;
-    } else {
-	char *arg = Tcl_GetStringFromObj(objv[1],NULL);
-	if ((argc>1) && (arg[0] == '-')
-		&& (arg[1] >= 'a') && (arg[1] <= 'z')) {
-	    i = 1;
-	} else {
-	    i = 2;
-	}
-    }
-
-    if (argc < i) {
-	Tcl_AppendResult(interp, "wrong # args: should be \"",
-		Tk_PathName(Tk_CanvasTkwin(canvas)), " create ",
-		itemPtr->typePtr->name, " x y ?options?\"", (char *) NULL);
-	return TCL_ERROR;
+    if (objc == 0) {
+	panic("canvas did not pass any coords\n");
     }
 
     /*
-     * Carry out initialization that is needed in order to clean
-     * up after errors during the the remainder of this procedure.
+     * Carry out initialization that is needed in order to clean up after
+     * errors during the the remainder of this procedure.
      */
 
     textPtr->textInfoPtr = Tk_CanvasGetTextInfo(canvas);
@@ -286,9 +261,6 @@ CreateText(interp, canvas, itemPtr, argc, argv)
     textPtr->insertPos	= 0;
 
     textPtr->anchor	= TK_ANCHOR_CENTER;
-    textPtr->tile = NULL;
-    textPtr->activeTile = NULL;
-    textPtr->disabledTile = NULL;
     textPtr->tsoffset.flags = 0;
     textPtr->tsoffset.xoffset = 0;
     textPtr->tsoffset.yoffset = 0;
@@ -304,21 +276,33 @@ CreateText(interp, canvas, itemPtr, argc, argv)
     textPtr->width	= 0;
 
     textPtr->numChars	= 0;
+    textPtr->numBytes	= 0;
     textPtr->textLayout = NULL;
     textPtr->leftEdge	= 0;
     textPtr->rightEdge	= 0;
     textPtr->gc		= None;
     textPtr->selTextGC	= None;
     textPtr->cursorOffGC = None;
+    textPtr->tile        = NULL;
 
     /*
      * Process the arguments to fill in the item record.
+     * Only 1 (list) or 2 (x y) coords are allowed.
      */
 
-    if ((TextCoords(interp, canvas, itemPtr, i, argv) != TCL_OK)) {
+    if (objc == 1) {
+	i = 1;
+    } else {
+	char *arg = Tcl_GetString(objv[1]);
+	i = 2;
+	if ((arg[0] == '-') && (arg[1] >= 'a') && (arg[1] <= 'z')) {
+	    i = 1;
+	}
+    }
+    if ((TextCoords(interp, canvas, itemPtr, i, objv) != TCL_OK)) {
 	goto error;
     }
-    if (ConfigureText(interp, canvas, itemPtr, argc-i, argv+i, 0) == TCL_OK) {
+    if (ConfigureText(interp, canvas, itemPtr, objc-i, objv+i, 0) == TCL_OK) {
 	return TCL_OK;
     }
 
@@ -326,7 +310,7 @@ CreateText(interp, canvas, itemPtr, argc, argv)
     DeleteText(canvas, itemPtr, Tk_Display(Tk_CanvasTkwin(canvas)));
     return TCL_ERROR;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -337,7 +321,7 @@ CreateText(interp, canvas, itemPtr, argc, argv)
  *	details on what it does.
  *
  * Results:
- *	Returns TCL_OK or TCL_ERROR, and sets interp->result.
+ *	Returns TCL_OK or TCL_ERROR, and sets the interp's result.
  *
  * Side effects:
  *	The coordinates for the given item may be changed.
@@ -346,34 +330,33 @@ CreateText(interp, canvas, itemPtr, argc, argv)
  */
 
 static int
-TextCoords(interp, canvas, itemPtr, argc, argv)
-    Tcl_Interp *interp;			/* Used for error reporting. */
-    Tk_Canvas canvas;			/* Canvas containing item. */
-    Tk_Item *itemPtr;			/* Item whose coordinates are to be
-					 * read or modified. */
-    int argc;				/* Number of coordinates supplied in
-					 * argv. */
-    char **argv;			/* Array of coordinates: x1, y1,
-					 * x2, y2, ... */
+TextCoords(interp, canvas, itemPtr, objc, objv)
+    Tcl_Interp *interp;		/* Used for error reporting. */
+    Tk_Canvas canvas;		/* Canvas containing item. */
+    Tk_Item *itemPtr;		/* Item whose coordinates are to be read or
+				 * modified. */
+    int objc;			/* Number of coordinates supplied in objv. */
+    Tcl_Obj *CONST objv[];	/* Array of coordinates: x1, y1, x2, y2, ... */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
-    char x[TCL_DOUBLE_SPACE];
 
-    if (argc == 0) {
+    if (objc == 0) {
 	Tcl_Obj *obj = Tcl_NewObj();
 	Tcl_Obj *subobj = Tcl_NewDoubleObj(textPtr->x);
 	Tcl_ListObjAppendElement(interp, obj, subobj);
 	subobj = Tcl_NewDoubleObj(textPtr->y);
 	Tcl_ListObjAppendElement(interp, obj, subobj);
 	Tcl_SetObjResult(interp, obj);
-    } else if (argc < 3) {
-	if (argc==1) {
-	    if (Tcl_ListObjGetElements(interp, objv[0], &argc, &objv) != TCL_OK) {
+    } else if (objc < 3) {
+	if (objc==1) {
+	    if (Tcl_ListObjGetElements(interp, objv[0], &objc,
+		    (Tcl_Obj ***) &objv) != TCL_OK) {
 		return TCL_ERROR;
-	    } else if (argc != 2) {
-		sprintf(x,"%d",argc);
-		Tcl_AppendResult(interp, "wrong # coordinates: expected 2, got ",
-		x, (char *) NULL);
+	    } else if (objc != 2) {
+		char buf[64 + TCL_INTEGER_SPACE];
+
+		sprintf(buf, "wrong # coordinates: expected 2, got %d", objc);
+		Tcl_SetResult(interp, buf, TCL_VOLATILE);
 		return TCL_ERROR;
 	    }
 	}
@@ -384,14 +367,15 @@ TextCoords(interp, canvas, itemPtr, argc, argv)
 	}
 	ComputeTextBbox(canvas, textPtr);
     } else {
-	sprintf(x,"%d",argc);
-	Tcl_AppendResult(interp, "wrong # coordinates: expected 0 or 2, got ",
-	x, (char *) NULL);
+	char buf[64 + TCL_INTEGER_SPACE];
+
+	sprintf(buf, "wrong # coordinates: expected 0 or 2, got %d", objc);
+	Tcl_SetResult(interp, buf, TCL_VOLATILE);
 	return TCL_ERROR;
     }
     return TCL_OK;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -402,7 +386,7 @@ TextCoords(interp, canvas, itemPtr, argc, argv)
  *
  * Results:
  *	A standard Tcl result code.  If an error occurs, then
- *	an error message is left in interp->result.
+ *	an error message is left in the interp's result.
  *
  * Side effects:
  *	Configuration information, such as colors and stipple
@@ -412,12 +396,12 @@ TextCoords(interp, canvas, itemPtr, argc, argv)
  */
 
 static int
-ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
+ConfigureText(interp, canvas, itemPtr, objc, objv, flags)
     Tcl_Interp *interp;		/* Interpreter for error reporting. */
     Tk_Canvas canvas;		/* Canvas containing itemPtr. */
     Tk_Item *itemPtr;		/* Rectangle item to reconfigure. */
-    int argc;			/* Number of elements in argv.  */
-    char **argv;		/* Arguments describing things to configure. */
+    int objc;			/* Number of elements in objv.  */
+    Tcl_Obj *CONST objv[];	/* Arguments describing things to configure. */
     int flags;			/* Flags to pass to Tk_ConfigureWidget. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
@@ -427,15 +411,14 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
     Tk_Window tkwin;
     Tk_CanvasTextInfo *textInfoPtr = textPtr->textInfoPtr;
     XColor *selBgColorPtr;
-    Tk_Tile tile;
     XColor *color;
     Pixmap stipple;
     Tk_State state;
-    Pixmap pixmap;
+    Tk_Tile tile;
 
     tkwin = Tk_CanvasTkwin(canvas);
-    if (Tk_ConfigureWidget(interp, tkwin, configSpecs, argc, argv,
-	    (char *) textPtr, flags|TK_CONFIG_OBJS) != TCL_OK) {
+    if (TCL_OK != Tk_ConfigureWidget(interp, tkwin, configSpecs, objc,
+	    (CONST84 Tcl_Obj *CONST*) objv, (char *) textPtr, flags|TK_CONFIG_OBJS)) {
 	return TCL_ERROR;
     }
 
@@ -447,7 +430,6 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
     state = Tk_GetItemState(canvas, itemPtr);
 
     if (textPtr->activeColor != NULL ||
-	    textPtr->activeTile != NULL ||
 	    textPtr->activeStipple != None) {
 	itemPtr->redraw_flags |= TK_ITEM_STATE_DEPENDANT;
     } else {
@@ -458,9 +440,6 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
     color = textPtr->color;
     stipple = textPtr->stipple;
     if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
-	if (textPtr->activeTile!=NULL) {
-	    tile = textPtr->activeTile;
-	}
 	if (textPtr->activeColor!=NULL) {
 	    color = textPtr->activeColor;
 	}
@@ -468,9 +447,6 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
 	    stipple = textPtr->activeStipple;
 	}
     } else if (state==TK_STATE_DISABLED) {
-	if (textPtr->disabledTile!=NULL) {
-	    tile = textPtr->disabledTile;
-	}
 	if (textPtr->disabledColor!=NULL) {
 	    color = textPtr->disabledColor;
 	}
@@ -479,20 +455,11 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
 	}
     }
 
-    Tk_SetTileCanvasItem(textPtr->tile, canvas, (Tk_Item *) NULL);
-    Tk_SetTileCanvasItem(textPtr->activeTile, canvas, (Tk_Item *) NULL);
-    Tk_SetTileCanvasItem(textPtr->disabledTile, canvas, (Tk_Item *) NULL);
-    Tk_SetTileCanvasItem(tile, canvas, itemPtr);
-
     newGC = newSelGC = None;
     if (textPtr->tkfont != NULL) {
 	gcValues.font = Tk_FontId(textPtr->tkfont);
 	mask = GCFont;
-	if ((pixmap = Tk_PixmapOfTile(tile)) != None) {
-	    gcValues.fill_style = FillTiled;
-	    gcValues.tile = pixmap;
-	    newGC = Tk_GetGC(tkwin, GCFont|GCTile|GCFillStyle, &gcValues);
-	} else if (color != NULL) {
+	if (color != NULL) {
 	    gcValues.foreground = color->pixel;
 	    mask |= GCForeground;
 	    if (stipple != None) {
@@ -543,17 +510,19 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
      * to keep them inside the item.
      */
 
-    textPtr->numChars = strlen(textPtr->text);
+    textPtr->numBytes = strlen(textPtr->text);
+    textPtr->numChars = Tcl_NumUtfChars(textPtr->text, textPtr->numBytes);
     if (textInfoPtr->selItemPtr == itemPtr) {
+
 	if (textInfoPtr->selectFirst >= textPtr->numChars) {
 	    textInfoPtr->selItemPtr = NULL;
 	} else {
 	    if (textInfoPtr->selectLast >= textPtr->numChars) {
-		textInfoPtr->selectLast = textPtr->numChars-1;
+		textInfoPtr->selectLast = textPtr->numChars - 1;
 	    }
 	    if ((textInfoPtr->anchorItemPtr == itemPtr)
 		    && (textInfoPtr->selectAnchor >= textPtr->numChars)) {
-		textInfoPtr->selectAnchor = textPtr->numChars-1;
+		textInfoPtr->selectAnchor = textPtr->numChars - 1;
 	    }
 	}
     }
@@ -564,7 +533,7 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
     ComputeTextBbox(canvas, textPtr);
     return TCL_OK;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -584,22 +553,12 @@ ConfigureText(interp, canvas, itemPtr, argc, argv, flags)
 
 static void
 DeleteText(canvas, itemPtr, display)
-    Tk_Canvas canvas;			/* Info about overall canvas widget. */
-    Tk_Item *itemPtr;			/* Item that is being deleted. */
-    Display *display;			/* Display containing window for
-					 * canvas. */
+    Tk_Canvas canvas;		/* Info about overall canvas widget. */
+    Tk_Item *itemPtr;		/* Item that is being deleted. */
+    Display *display;		/* Display containing window for canvas. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
 
-    if (textPtr->tile != NULL) {
-	Tk_FreeTile(textPtr->tile);
-    }
-    if (textPtr->activeTile != NULL) {
-	Tk_FreeTile(textPtr->activeTile);
-    }
-    if (textPtr->disabledTile != NULL) {
-	Tk_FreeTile(textPtr->disabledTile);
-    }
     if (textPtr->color != NULL) {
 	Tk_FreeColor(textPtr->color);
     }
@@ -634,7 +593,7 @@ DeleteText(canvas, itemPtr, display)
 	Tk_FreeGC(display, textPtr->cursorOffGC);
     }
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -658,9 +617,8 @@ DeleteText(canvas, itemPtr, display)
 
 static void
 ComputeTextBbox(canvas, textPtr)
-    Tk_Canvas canvas;			/* Canvas that contains item. */
-    TextItem *textPtr;			/* Item whose bbos is to be
-					 * recomputed. */
+    Tk_Canvas canvas;		/* Canvas that contains item. */
+    TextItem *textPtr;		/* Item whose bbox is to be recomputed. */
 {
     Tk_CanvasTextInfo *textInfoPtr;
     int leftX, topY, width, height, fudge;
@@ -680,8 +638,8 @@ ComputeTextBbox(canvas, textPtr)
      * of the bounding box for the text item.
      */
 
-    leftX = (int) (textPtr->x + 0.5);
-    topY = (int) (textPtr->y + 0.5);
+    leftX = (int) floor(textPtr->x + 0.5);
+    topY = (int) floor(textPtr->y + 0.5);
     switch (textPtr->anchor) {
 	case TK_ANCHOR_NW:
 	case TK_ANCHOR_N:
@@ -739,7 +697,7 @@ ComputeTextBbox(canvas, textPtr)
     textPtr->header.x2 = leftX + width + fudge;
     textPtr->header.y2 = topY + height;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -760,17 +718,16 @@ ComputeTextBbox(canvas, textPtr)
 
 static void
 DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
-    Tk_Canvas canvas;			/* Canvas that contains item. */
-    Tk_Item *itemPtr;			/* Item to be displayed. */
-    Display *display;			/* Display on which to draw item. */
-    Drawable drawable;			/* Pixmap or window in which to draw
-					 * item. */
-    int x, y, width, height;		/* Describes region of canvas that
-					 * must be redisplayed (not used). */
+    Tk_Canvas canvas;		/* Canvas that contains item. */
+    Tk_Item *itemPtr;		/* Item to be displayed. */
+    Display *display;		/* Display on which to draw item. */
+    Drawable drawable;		/* Pixmap or window in which to draw item. */
+    int x, y, width, height;	/* Describes region of canvas that must be
+				 * redisplayed (not used). */
 {
     TextItem *textPtr;
     Tk_CanvasTextInfo *textInfoPtr;
-    int selFirst, selLast;
+    int selFirstChar, selLastChar;
     short drawableX, drawableY;
     Pixmap stipple;
     Tk_Tile tile;
@@ -780,21 +737,14 @@ DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
     textInfoPtr = textPtr->textInfoPtr;
 
     stipple = textPtr->stipple;
-    tile = textPtr->tile;
     if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
 	if (textPtr->activeStipple!=None) {
 	    stipple = textPtr->activeStipple;
 	}
-	if (textPtr->activeTile!=NULL) {
-	    tile = textPtr->activeTile;
-	}
     } else if (state==TK_STATE_DISABLED) {
 	if (textPtr->disabledStipple!=None) {
 	    stipple = textPtr->disabledStipple;
 	}
-	if (textPtr->disabledTile!=NULL) {
-	    tile = textPtr->disabledTile;
-	}
     }
 
     if (textPtr->gc == None) {
@@ -807,30 +757,31 @@ DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
      * read-only.
      */
 
-    if ((tile != NULL) || (stipple != None)) {
+    if (stipple != None) {
 	Tk_CanvasSetOffset(canvas, textPtr->gc, &textPtr->tsoffset);
     }
 
-    selFirst = -1;
-    selLast = 0;		/* lint. */
+    selFirstChar = -1;
+    selLastChar = 0;		/* lint. */
+
     if (textInfoPtr->selItemPtr == itemPtr) {
-	selFirst = textInfoPtr->selectFirst;
-	selLast = textInfoPtr->selectLast;
-	if (selLast >= textPtr->numChars) {
-	    selLast = textPtr->numChars - 1;
+	selFirstChar = textInfoPtr->selectFirst;
+	selLastChar = textInfoPtr->selectLast;
+	if (selLastChar > textPtr->numChars) {
+	    selLastChar = textPtr->numChars - 1;
 	}
-	if ((selFirst >= 0) && (selFirst <= selLast)) {
+	if ((selFirstChar >= 0) && (selFirstChar <= selLastChar)) {
+	    int xFirst, yFirst, hFirst;
+	    int xLast, yLast, wLast;
+
 	    /*
 	     * Draw a special background under the selection.
 	     */
 
-	    int xFirst, yFirst, hFirst;
-	    int xLast, yLast, wLast;
-
-	    Tk_CharBbox(textPtr->textLayout, selFirst,
-		    &xFirst, &yFirst, NULL, &hFirst);
-	    Tk_CharBbox(textPtr->textLayout, selLast,
-		    &xLast, &yLast, &wLast, NULL);
+	    Tk_CharBbox(textPtr->textLayout, selFirstChar, &xFirst, &yFirst,
+		    NULL, &hFirst);
+	    Tk_CharBbox(textPtr->textLayout, selLastChar, &xLast, &yLast,
+		    &wLast, NULL);
 
 	    /*
 	     * If the selection spans the end of this line, then display
@@ -843,8 +794,8 @@ DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
 	    height = hFirst;
 	    for (y = yFirst ; y <= yLast; y += height) {
 		if (y == yLast) {
-		    width = (xLast + wLast) - x;
-		} else {	
+		    width = xLast + wLast - x;
+		} else {
 		    width = textPtr->rightEdge - textPtr->leftEdge - x;
 		}
 		Tk_CanvasDrawableCoords(canvas,
@@ -878,6 +829,8 @@ DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
 			    - (textInfoPtr->insertWidth / 2)),
 		    (double) (textPtr->header.y1 + y),
 		    &drawableX, &drawableY);
+	    Tk_SetCaretPos(Tk_CanvasTkwin(canvas), drawableX, drawableY,
+		    height);
 	    if (textInfoPtr->cursorOn) {
 		Tk_Fill3DRectangle(Tk_CanvasTkwin(canvas), drawable,
 			textInfoPtr->insertBorder,
@@ -914,17 +867,17 @@ DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
     Tk_DrawTextLayout(display, drawable, textPtr->gc, textPtr->textLayout,
 	    drawableX, drawableY, 0, -1);
 
-    if ((selFirst >= 0) && (textPtr->selTextGC != textPtr->gc)) {
+    if ((selFirstChar >= 0) && (textPtr->selTextGC != textPtr->gc)) {
 	Tk_DrawTextLayout(display, drawable, textPtr->selTextGC,
-	    textPtr->textLayout, drawableX, drawableY, selFirst,
-	    selLast + 1);
+	    textPtr->textLayout, drawableX, drawableY, selFirstChar,
+	    selLastChar + 1);
     }
 
-    if ((tile != NULL) ||(stipple != None)) {
+    if (stipple != None) {
 	XSetTSOrigin(display, textPtr->gc, 0, 0);
     }
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -944,37 +897,44 @@ DisplayCanvText(canvas, itemPtr, display, drawable, x, y, width, height)
  */
 
 static void
-TextInsert(canvas, itemPtr, beforeThis, ostring)
+TextInsert(canvas, itemPtr, index, ostring)
     Tk_Canvas canvas;		/* Canvas containing text item. */
     Tk_Item *itemPtr;		/* Text item to be modified. */
-    int beforeThis;		/* Index of character before which text is
+    int index;			/* Character index before which string is
 				 * to be inserted. */
     Tcl_Obj *ostring;		/* New characters to be inserted. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
-    int length = 0;
-    char *new;
+    int byteIndex, byteCount, charsAdded;
+    char *new, *text;
     Tk_CanvasTextInfo *textInfoPtr = textPtr->textInfoPtr;
 
-    char * string = Tcl_GetStringFromObj(ostring, &length);
+    char * string = Tcl_GetString(ostring);
 
-    if (length == 0) {
+    text = textPtr->text;
+
+    if (index < 0) {
+	index = 0;
+    }
+    if (index > textPtr->numChars) {
+	index = textPtr->numChars;
+    }
+    byteIndex = Tcl_UtfAtIndex(text, index) - text;
+    byteCount = strlen(string);
+    if (byteCount == 0) {
 	return;
     }
-    if (beforeThis < 0) {
-	beforeThis = 0;
-    }
-    if (beforeThis > textPtr->numChars) {
-	beforeThis = textPtr->numChars;
-    }
 
-    new = (char *) ckalloc((unsigned) (textPtr->numChars + length + 1));
-    strncpy(new, textPtr->text, (size_t) beforeThis);
-    strcpy(new+beforeThis, string);
-    strcpy(new+beforeThis+length, textPtr->text+beforeThis);
-    ckfree(textPtr->text);
+    new = (char *) ckalloc((unsigned) textPtr->numBytes + byteCount + 1);
+    memcpy(new, text, (size_t) byteIndex);
+    strcpy(new + byteIndex, string);
+    strcpy(new + byteIndex + byteCount, text + byteIndex);
+
+    ckfree(text);
     textPtr->text = new;
-    textPtr->numChars += length;
+    charsAdded = Tcl_NumUtfChars(string, byteCount);
+    textPtr->numChars += charsAdded;
+    textPtr->numBytes += byteCount;
 
     /*
      * Inserting characters invalidates indices such as those for the
@@ -982,23 +942,23 @@ TextInsert(canvas, itemPtr, beforeThis, ostring)
      */
 
     if (textInfoPtr->selItemPtr == itemPtr) {
-	if (textInfoPtr->selectFirst >= beforeThis) {
-	    textInfoPtr->selectFirst += length;
+	if (textInfoPtr->selectFirst >= index) {
+	    textInfoPtr->selectFirst += charsAdded;
 	}
-	if (textInfoPtr->selectLast >= beforeThis) {
-	    textInfoPtr->selectLast += length;
+	if (textInfoPtr->selectLast >= index) {
+	    textInfoPtr->selectLast += charsAdded;
 	}
 	if ((textInfoPtr->anchorItemPtr == itemPtr)
-		&& (textInfoPtr->selectAnchor >= beforeThis)) {
-	    textInfoPtr->selectAnchor += length;
+		&& (textInfoPtr->selectAnchor >= index)) {
+	    textInfoPtr->selectAnchor += charsAdded;
 	}
     }
-    if (textPtr->insertPos >= beforeThis) {
-	textPtr->insertPos += length;
+    if (textPtr->insertPos >= index) {
+	textPtr->insertPos += charsAdded;
     }
     ComputeTextBbox(canvas, textPtr);
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1021,31 +981,40 @@ static void
 TextDeleteChars(canvas, itemPtr, first, last)
     Tk_Canvas canvas;		/* Canvas containing itemPtr. */
     Tk_Item *itemPtr;		/* Item in which to delete characters. */
-    int first;			/* Index of first character to delete. */
-    int last;			/* Index of last character to delete. */
+    int first;			/* Character index of first character to
+				 * delete. */
+    int last;			/* Character index of last character to
+				 * delete (inclusive). */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
-    int count;
-    char *new;
+    int byteIndex, byteCount, charsRemoved;
+    char *new, *text;
     Tk_CanvasTextInfo *textInfoPtr = textPtr->textInfoPtr;
 
+    text = textPtr->text;
     if (first < 0) {
 	first = 0;
     }
     if (last >= textPtr->numChars) {
-	last = textPtr->numChars-1;
+	last = textPtr->numChars - 1;
     }
     if (first > last) {
 	return;
     }
-    count = last + 1 - first;
+    charsRemoved = last + 1 - first;
 
-    new = (char *) ckalloc((unsigned) (textPtr->numChars + 1 - count));
-    strncpy(new, textPtr->text, (size_t) first);
-    strcpy(new+first, textPtr->text+last+1);
-    ckfree(textPtr->text);
+    byteIndex = Tcl_UtfAtIndex(text, first) - text;
+    byteCount = Tcl_UtfAtIndex(text + byteIndex, charsRemoved)
+	- (text + byteIndex);
+
+    new = (char *) ckalloc((unsigned) (textPtr->numBytes + 1 - byteCount));
+    memcpy(new, text, (size_t) byteIndex);
+    strcpy(new + byteIndex, text + byteIndex + byteCount);
+
+    ckfree(text);
     textPtr->text = new;
-    textPtr->numChars -= count;
+    textPtr->numChars -= charsRemoved;
+    textPtr->numBytes -= byteCount;
 
     /*
      * Update indexes for the selection and cursor to reflect the
@@ -1054,15 +1023,15 @@ TextDeleteChars(canvas, itemPtr, first, last)
 
     if (textInfoPtr->selItemPtr == itemPtr) {
 	if (textInfoPtr->selectFirst > first) {
-	    textInfoPtr->selectFirst -= count;
+	    textInfoPtr->selectFirst -= charsRemoved;
 	    if (textInfoPtr->selectFirst < first) {
 		textInfoPtr->selectFirst = first;
 	    }
 	}
 	if (textInfoPtr->selectLast >= first) {
-	    textInfoPtr->selectLast -= count;
-	    if (textInfoPtr->selectLast < (first-1)) {
-		textInfoPtr->selectLast = (first-1);
+	    textInfoPtr->selectLast -= charsRemoved;
+	    if (textInfoPtr->selectLast < first - 1) {
+		textInfoPtr->selectLast = first - 1;
 	    }
 	}
 	if (textInfoPtr->selectFirst > textInfoPtr->selectLast) {
@@ -1070,14 +1039,14 @@ TextDeleteChars(canvas, itemPtr, first, last)
 	}
 	if ((textInfoPtr->anchorItemPtr == itemPtr)
 		&& (textInfoPtr->selectAnchor > first)) {
-	    textInfoPtr->selectAnchor -= count;
+	    textInfoPtr->selectAnchor -= charsRemoved;
 	    if (textInfoPtr->selectAnchor < first) {
 		textInfoPtr->selectAnchor = first;
 	    }
 	}
     }
     if (textPtr->insertPos > first) {
-	textPtr->insertPos -= count;
+	textPtr->insertPos -= charsRemoved;
 	if (textPtr->insertPos < first) {
 	    textPtr->insertPos = first;
 	}
@@ -1085,7 +1054,7 @@ TextDeleteChars(canvas, itemPtr, first, last)
     ComputeTextBbox(canvas, textPtr);
     return;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1127,7 +1096,7 @@ TextToPoint(canvas, itemPtr, pointPtr)
     }
     return value;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1166,7 +1135,7 @@ TextToArea(canvas, itemPtr, rectPtr)
 	    (int) (rectPtr[2] - rectPtr[0] + 0.5),
 	    (int) (rectPtr[3] - rectPtr[1] + 0.5));
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1187,11 +1156,11 @@ TextToArea(canvas, itemPtr, rectPtr)
 	/* ARGSUSED */
 static void
 ScaleText(canvas, itemPtr, originX, originY, scaleX, scaleY)
-    Tk_Canvas canvas;			/* Canvas containing rectangle. */
-    Tk_Item *itemPtr;			/* Rectangle to be scaled. */
-    double originX, originY;		/* Origin about which to scale rect. */
-    double scaleX;			/* Amount to scale in X direction. */
-    double scaleY;			/* Amount to scale in Y direction. */
+    Tk_Canvas canvas;		/* Canvas containing rectangle. */
+    Tk_Item *itemPtr;		/* Rectangle to be scaled. */
+    double originX, originY;	/* Origin about which to scale rect. */
+    double scaleX;		/* Amount to scale in X direction. */
+    double scaleY;		/* Amount to scale in Y direction. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
 
@@ -1200,7 +1169,7 @@ ScaleText(canvas, itemPtr, originX, originY, scaleX, scaleY)
     ComputeTextBbox(canvas, textPtr);
     return;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1222,10 +1191,9 @@ ScaleText(canvas, itemPtr, originX, originY, scaleX, scaleY)
 
 static void
 TranslateText(canvas, itemPtr, deltaX, deltaY)
-    Tk_Canvas canvas;			/* Canvas containing item. */
-    Tk_Item *itemPtr;			/* Item that is being moved. */
-    double deltaX, deltaY;		/* Amount by which item is to be
-					 * moved. */
+    Tk_Canvas canvas;		/* Canvas containing item. */
+    Tk_Item *itemPtr;		/* Item that is being moved. */
+    double deltaX, deltaY;	/* Amount by which item is to be moved. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
 
@@ -1233,7 +1201,7 @@ TranslateText(canvas, itemPtr, deltaX, deltaY)
     textPtr->y += deltaY;
     ComputeTextBbox(canvas, textPtr);
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1246,7 +1214,7 @@ TranslateText(canvas, itemPtr, deltaX, deltaY)
  *	A standard Tcl result.  If all went well, then *indexPtr is
  *	filled in with the index (into itemPtr) corresponding to
  *	string.  Otherwise an error message is left in
- *	interp->result.
+ *	the interp's result.
  *
  * Side effects:
  *	None.
@@ -1262,10 +1230,11 @@ GetTextIndex(interp, canvas, itemPtr, obj, indexPtr)
 				 * specified. */
     Tcl_Obj *obj;		/* Specification of a particular character
 				 * in itemPtr's text. */
-    int *indexPtr;		/* Where to store converted index. */
+    int *indexPtr;		/* Where to store converted character
+				 * index. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
-    int length;
+    size_t length;
     int c;
     TkCanvas *canvasPtr = (TkCanvas *) canvas;
     Tk_CanvasTextInfo *textInfoPtr = textPtr->textInfoPtr;
@@ -1283,6 +1252,7 @@ GetTextIndex(interp, canvas, itemPtr, obj, indexPtr)
 
     string = Tcl_GetStringFromObj(obj, &length);
     c = string[0];
+    length = strlen(string);
 
     if ((c == 'e') && (strncmp(string, "end", length) == 0)) {
 	*indexPtr = textPtr->numChars;
@@ -1291,14 +1261,14 @@ GetTextIndex(interp, canvas, itemPtr, obj, indexPtr)
     } else if ((c == 's') && (strncmp(string, "sel.first", length) == 0)
 	    && (length >= 5)) {
 	if (textInfoPtr->selItemPtr != itemPtr) {
-	    Tcl_AppendResult(interp, "selection isn't in item", (char *) NULL);
+	    Tcl_SetResult(interp, "selection isn't in item", TCL_STATIC);
 	    return TCL_ERROR;
 	}
 	*indexPtr = textInfoPtr->selectFirst;
     } else if ((c == 's') && (strncmp(string, "sel.last", length) == 0)
 	    && (length >= 5)) {
 	if (textInfoPtr->selItemPtr != itemPtr) {
-	    Tcl_AppendResult(interp, "selection isn't in item", (char *) NULL);
+	    Tcl_SetResult(interp, "selection isn't in item", TCL_STATIC);
 	    return TCL_ERROR;
 	}
 	*indexPtr = textInfoPtr->selectLast;
@@ -1326,14 +1296,20 @@ GetTextIndex(interp, canvas, itemPtr, obj, indexPtr)
 	    *indexPtr = textPtr->numChars;
 	}
     } else {
+	/*
+	 * Some of the paths here leave messages in the interp's result,
+	 * so we have to clear it out before storing our own message.
+	 */
+
 	badIndex:
+	Tcl_SetResult(interp, (char *) NULL, TCL_STATIC);
 	Tcl_AppendResult(interp, "bad index \"", string, "\"",
 		(char *) NULL);
 	return TCL_ERROR;
     }
     return TCL_OK;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1353,11 +1329,11 @@ GetTextIndex(interp, canvas, itemPtr, obj, indexPtr)
 	/* ARGSUSED */
 static void
 SetTextCursor(canvas, itemPtr, index)
-    Tk_Canvas canvas;			/* Record describing canvas widget. */
-    Tk_Item *itemPtr;			/* Text item in which cursor position
-					 * is to be set. */
-    int index;				/* Index of character just before which
-					 * cursor is to be positioned. */
+    Tk_Canvas canvas;		/* Record describing canvas widget. */
+    Tk_Item *itemPtr;		/* Text item in which cursor position is to
+				 * be set. */
+    int index;			/* Character index of character just before
+				 * which cursor is to be positioned. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
 
@@ -1369,7 +1345,7 @@ SetTextCursor(canvas, itemPtr, index)
 	textPtr->insertPos = index;
     }
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1393,36 +1369,41 @@ SetTextCursor(canvas, itemPtr, index)
 
 static int
 GetSelText(canvas, itemPtr, offset, buffer, maxBytes)
-    Tk_Canvas canvas;			/* Canvas containing selection. */
-    Tk_Item *itemPtr;			/* Text item containing selection. */
-    int offset;				/* Offset within selection of first
-					 * character to be returned. */
-    char *buffer;			/* Location in which to place
-					 * selection. */
-    int maxBytes;			/* Maximum number of bytes to place
-					 * at buffer, not including terminating
-					 * NULL character. */
+    Tk_Canvas canvas;		/* Canvas containing selection. */
+    Tk_Item *itemPtr;		/* Text item containing selection. */
+    int offset;			/* Byte offset within selection of first
+				 * character to be returned. */
+    char *buffer;		/* Location in which to place selection. */
+    int maxBytes;		/* Maximum number of bytes to place at
+				 * buffer, not including terminating NULL
+				 * character. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
-    int count;
+    int byteCount;
+    char *text;
+    CONST char *selStart, *selEnd;
     Tk_CanvasTextInfo *textInfoPtr = textPtr->textInfoPtr;
 
-    count = textInfoPtr->selectLast + 1 - textInfoPtr->selectFirst - offset;
-    if (textInfoPtr->selectLast == textPtr->numChars) {
-	count -= 1;
-    }
-    if (count > maxBytes) {
-	count = maxBytes;
-    }
-    if (count <= 0) {
+    if ((textInfoPtr->selectFirst < 0) ||
+	    (textInfoPtr->selectFirst > textInfoPtr->selectLast)) {
 	return 0;
     }
-    strncpy(buffer, textPtr->text + textInfoPtr->selectFirst + offset,
-	    (size_t) count);
-    buffer[count] = '\0';
-    return count;
+    text = textPtr->text;
+    selStart = Tcl_UtfAtIndex(text, textInfoPtr->selectFirst);
+    selEnd = Tcl_UtfAtIndex(selStart,
+	    textInfoPtr->selectLast + 1 - textInfoPtr->selectFirst);
+    byteCount = selEnd - selStart - offset;
+    if (byteCount > maxBytes) {
+	byteCount = maxBytes;
+    }
+    if (byteCount <= 0) {
+	return 0;
+    }
+    memcpy(buffer, selStart + offset, (size_t) byteCount);
+    buffer[byteCount] = '\0';
+    return byteCount;
 }
-
+
 /*
  *--------------------------------------------------------------
  *
@@ -1434,7 +1415,7 @@ GetSelText(canvas, itemPtr, offset, buffer, maxBytes)
  * Results:
  *	The return value is a standard Tcl result.  If an error
  *	occurs in generating Postscript then an error message is
- *	left in interp->result, replacing whatever used
+ *	left in the interp's result, replacing whatever used
  *	to be there.  If no error occurs, then Postscript for the
  *	item is appended to the result.
  *
@@ -1446,14 +1427,12 @@ GetSelText(canvas, itemPtr, offset, buffer, maxBytes)
 
 static int
 TextToPostscript(interp, canvas, itemPtr, prepass)
-    Tcl_Interp *interp;			/* Leave Postscript or error message
-					 * here. */
-    Tk_Canvas canvas;			/* Information about overall canvas. */
-    Tk_Item *itemPtr;			/* Item for which Postscript is
-					 * wanted. */
-    int prepass;			/* 1 means this is a prepass to
-					 * collect font information;  0 means
-					 * final Postscript is being created. */
+    Tcl_Interp *interp;		/* Leave Postscript or error message here. */
+    Tk_Canvas canvas;		/* Information about overall canvas. */
+    Tk_Item *itemPtr;		/* Item for which Postscript is wanted. */
+    int prepass;		/* 1 means this is a prepass to collect
+				 * font information; 0 means final Postscript
+				 * is being created. */
 {
     TextItem *textPtr = (TextItem *) itemPtr;
     int x, y;
@@ -1533,3 +1512,5 @@ TextToPostscript(interp, canvas, itemPtr, prepass)
 
     return TCL_OK;
 }
+
+
